@@ -29,13 +29,12 @@
 # ===================================================================
 
 import sys
+import struct
 
-from Crypto.Util.py3compat import tobytes, is_native_int
+from Crypto.Util.py3compat import is_native_int
 
 from Crypto.Util._raw_api import (backend, load_lib,
-                                  get_raw_buffer, get_c_string,
-                                  null_pointer, create_string_buffer,
-                                  c_ulong, c_size_t)
+                                  c_ulong, c_size_t, c_uint8_ptr)
 
 from ._IntegerBase import IntegerBase
 
@@ -43,12 +42,14 @@ gmp_defs = """typedef unsigned long UNIX_ULONG;
         typedef struct { int a; int b; void *c; } MPZ;
         typedef MPZ mpz_t[1];
         typedef UNIX_ULONG mp_bitcnt_t;
+
         void __gmpz_init (mpz_t x);
         void __gmpz_init_set (mpz_t rop, const mpz_t op);
         void __gmpz_init_set_ui (mpz_t rop, UNIX_ULONG op);
-        int __gmp_sscanf (const char *s, const char *fmt, ...);
+
+        UNIX_ULONG __gmpz_get_ui (const mpz_t op);
         void __gmpz_set (mpz_t rop, const mpz_t op);
-        int __gmp_snprintf (uint8_t *buf, size_t size, const char *fmt, ...);
+        void __gmpz_set_ui (mpz_t rop, UNIX_ULONG op);
         void __gmpz_add (mpz_t rop, const mpz_t op1, const mpz_t op2);
         void __gmpz_add_ui (mpz_t rop, const mpz_t op1, UNIX_ULONG op2);
         void __gmpz_sub_ui (mpz_t rop, const mpz_t op1, UNIX_ULONG op2);
@@ -90,6 +91,9 @@ gmp_defs = """typedef unsigned long UNIX_ULONG;
         int __gmpz_invert (mpz_t rop, const mpz_t op1, const mpz_t op2);
         int __gmpz_divisible_p (const mpz_t n, const mpz_t d);
         int __gmpz_divisible_ui_p (const mpz_t n, UNIX_ULONG d);
+
+        size_t __gmpz_size (const mpz_t op);
+        UNIX_ULONG __gmpz_getlimbn (const mpz_t op, size_t n);
         """
 
 if sys.platform == "win32":
@@ -101,27 +105,6 @@ implementation = {"library": "gmp", "api": backend}
 if hasattr(lib, "__mpir_version"):
     raise ImportError("MPIR library detected")
 
-# In order to create a function that returns a pointer to
-# a new MPZ structure, we need to break the abstraction
-# and know exactly what ffi backend we have
-if implementation["api"] == "ctypes":
-    from ctypes import Structure, c_int, c_void_p, byref
-
-    class _MPZ(Structure):
-        _fields_ = [('_mp_alloc', c_int),
-                    ('_mp_size', c_int),
-                    ('_mp_d', c_void_p)]
-
-    def new_mpz():
-        return byref(_MPZ())
-
-else:
-    # We are using CFFI
-    from Crypto.Util._raw_api import ffi
-
-    def new_mpz():
-        return ffi.new("MPZ*")
-
 
 # Lazy creation of GMP methods
 class _GMP(object):
@@ -141,6 +124,34 @@ class _GMP(object):
 _gmp = _GMP()
 
 
+# In order to create a function that returns a pointer to
+# a new MPZ structure, we need to break the abstraction
+# and know exactly what ffi backend we have
+if implementation["api"] == "ctypes":
+    from ctypes import Structure, c_int, c_void_p, byref
+
+    class _MPZ(Structure):
+        _fields_ = [('_mp_alloc', c_int),
+                    ('_mp_size', c_int),
+                    ('_mp_d', c_void_p)]
+
+    def new_mpz():
+        return byref(_MPZ())
+
+    _gmp.mpz_getlimbn.restype = c_ulong
+
+else:
+    # We are using CFFI
+    from Crypto.Util._raw_api import ffi
+
+    def new_mpz():
+        return ffi.new("MPZ*")
+
+
+# Size of a native word
+_sys_bits = 8 * struct.calcsize("P")
+
+
 class IntegerGMP(IntegerBase):
     """A fast, arbitrary precision integer"""
 
@@ -156,28 +167,57 @@ class IntegerGMP(IntegerBase):
         if isinstance(value, float):
             raise ValueError("A floating point type is not a natural number")
 
-        self._initialized = True
-        
         if is_native_int(value):
             _gmp.mpz_init(self._mpz_p)
-            result = _gmp.gmp_sscanf(tobytes(str(value)), b"%Zd", self._mpz_p)
-            if result != 1:
-                raise ValueError("Error converting '%d'" % value)
+            self._initialized = True
+            if value == 0:
+                return
+
+            tmp = new_mpz()
+            _gmp.mpz_init(tmp)
+
+            try:
+                positive = value >= 0
+                reduce = abs(value)
+                slots = (reduce.bit_length() - 1) // 32 + 1
+
+                while slots > 0:
+                    slots = slots - 1
+                    _gmp.mpz_set_ui(tmp,
+                                    c_ulong(0xFFFFFFFF & (reduce >> (slots * 32))))
+                    _gmp.mpz_mul_2exp(tmp, tmp, c_ulong(slots * 32))
+                    _gmp.mpz_add(self._mpz_p, self._mpz_p, tmp)
+            finally:
+                _gmp.mpz_clear(tmp)
+
+            if not positive:
+                _gmp.mpz_neg(self._mpz_p, self._mpz_p)
+
         elif isinstance(value, IntegerGMP):
             _gmp.mpz_init_set(self._mpz_p, value._mpz_p)
+            self._initialized = True
         else:
             raise NotImplementedError
 
     # Conversions
     def __int__(self):
-        # buf will contain the integer encoded in decimal plus the trailing
-        # zero, and possibly the negative sign.
-        # dig10(x) < log10(x) + 1 = log2(x)/log2(10) + 1 < log2(x)/3 + 1
-        buf_len = _gmp.mpz_sizeinbase(self._mpz_p, 2) // 3 + 3
-        buf = create_string_buffer(buf_len)
+        tmp = new_mpz()
+        _gmp.mpz_init_set(tmp, self._mpz_p)
 
-        _gmp.gmp_snprintf(buf, c_size_t(buf_len), b"%Zd", self._mpz_p)
-        return int(get_c_string(buf))
+        try:
+            value = 0
+            slot = 0
+            while _gmp.mpz_cmp(tmp, self._zero_mpz_p) != 0:
+                lsb = _gmp.mpz_get_ui(tmp) & 0xFFFFFFFF
+                value |= lsb << (slot * 32)
+                _gmp.mpz_tdiv_q_2exp(tmp, tmp, c_ulong(32))
+                slot = slot + 1
+        finally:
+            _gmp.mpz_clear(tmp)
+
+        if self < 0:
+            value = -value
+        return int(value)
 
     def __str__(self):
         return str(int(self))
@@ -193,7 +233,7 @@ class IntegerGMP(IntegerBase):
     def __index__(self):
         return int(self)
 
-    def to_bytes(self, block_size=0):
+    def to_bytes(self, block_size=0, byteorder='big'):
         """Convert the number into a byte string.
 
         This method encodes the number in network order and prepends
@@ -204,6 +244,8 @@ class IntegerGMP(IntegerBase):
           block_size : integer
             The exact size the output byte string must have.
             If zero, the string has the minimal length.
+          byteorder : string
+            'big' for big-endian integers (default), 'little' for litte-endian.
         :Returns:
           A byte string.
         :Raise ValueError:
@@ -214,35 +256,65 @@ class IntegerGMP(IntegerBase):
         if self < 0:
             raise ValueError("Conversion only valid for non-negative numbers")
 
-        buf_len = (_gmp.mpz_sizeinbase(self._mpz_p, 2) + 7) // 8
-        if buf_len > block_size > 0:
-            raise ValueError("Number is too big to convert to byte string"
-                             "of prescribed length")
-        buf = create_string_buffer(buf_len)
+        num_limbs = _gmp.mpz_size(self._mpz_p)
+        if _sys_bits == 32:
+            spchar = "L"
+            num_limbs = max(1, num_limbs, (block_size + 3) // 4)
+        elif _sys_bits == 64:
+            spchar = "Q"
+            num_limbs = max(1, num_limbs, (block_size + 7) // 8)
+        else:
+            raise ValueError("Unknown limb size")
 
-        _gmp.mpz_export(
-                buf,
-                null_pointer,  # Ignore countp
-                1,             # Big endian
-                c_size_t(1),   # Each word is 1 byte long
-                0,             # Endianess within a word - not relevant
-                c_size_t(0),   # No nails
-                self._mpz_p)
+        # mpz_getlimbn returns 0 if i is larger than the number of actual limbs
+        limbs = [_gmp.mpz_getlimbn(self._mpz_p, num_limbs - i - 1) for i in range(num_limbs)]
 
-        return b'\x00' * max(0, block_size - buf_len) + get_raw_buffer(buf)
+        result = struct.pack(">" + spchar * num_limbs, *limbs)
+        cutoff_len = len(result) - block_size
+        if block_size == 0:
+            result = result.lstrip(b'\x00')
+        elif cutoff_len > 0:
+            if result[:cutoff_len] != b'\x00' * (cutoff_len):
+                raise ValueError("Number is too big to convert to "
+                                 "byte string of prescribed length")
+            result = result[cutoff_len:]
+        elif cutoff_len < 0:
+            result = b'\x00' * (-cutoff_len) + result
+
+        if byteorder == 'little':
+            result = result[::-1]
+        elif byteorder == 'big':
+            pass
+        else:
+            raise ValueError("Incorrect byteorder")
+
+        if len(result) == 0:
+            result = b'\x00'
+
+        return result
 
     @staticmethod
-    def from_bytes(byte_string):
+    def from_bytes(byte_string, byteorder='big'):
         """Convert a byte string into a number.
 
         :Parameters:
           byte_string : byte string
             The input number, encoded in network order.
             It can only be non-negative.
+          byteorder : string
+            'big' for big-endian integers (default), 'little' for litte-endian.
+
         :Return:
           The ``Integer`` object carrying the same value as the input.
         """
         result = IntegerGMP(0)
+        if byteorder == 'big':
+            pass
+        elif byteorder == 'little':
+            byte_string = bytearray(byte_string)
+            byte_string.reverse()
+        else:
+            raise ValueError("Incorrect byteorder")
         _gmp.mpz_import(
                         result._mpz_p,
                         c_size_t(len(byte_string)),  # Amount of words to read
@@ -250,7 +322,7 @@ class IntegerGMP(IntegerBase):
                         c_size_t(1),  # Each word is 1 byte long
                         0,            # Endianess within a word - not relevant
                         c_size_t(0),  # No nails
-                        byte_string)
+                        c_uint8_ptr(byte_string))
         return result
 
     # Relations
@@ -692,9 +764,28 @@ class IntegerGMP(IntegerBase):
         if not isinstance(n, IntegerGMP):
             n = IntegerGMP(n)
         if n <= 0 or n.is_even():
-            raise ValueError("n must be positive even for the Jacobi symbol")
+            raise ValueError("n must be positive odd for the Jacobi symbol")
         return _gmp.mpz_jacobi(a._mpz_p, n._mpz_p)
 
+    @staticmethod
+    def _mult_modulo_bytes(term1, term2, modulus):
+        if not isinstance(term1, IntegerGMP):
+            term1 = IntegerGMP(term1)
+        if not isinstance(term2, IntegerGMP):
+            term2 = IntegerGMP(term2)
+        if not isinstance(modulus, IntegerGMP):
+            modulus = IntegerGMP(modulus)
+
+        if modulus < 0:
+            raise ValueError("Modulus must be positive")
+        if modulus == 0:
+            raise ZeroDivisionError("Modulus cannot be zero")
+        if (modulus & 1) == 0:
+            raise ValueError("Odd modulus is required")
+
+        product = (term1 * term2) % modulus
+        return product.to_bytes(modulus.size_in_bytes())
+
     # Clean-up
     def __del__(self):
 
