@@ -1,4 +1,4 @@
-# -*- coding: ascii -*-
+# -*- coding: utf-8 -*-
 #
 #  Util/asn1.py : Minimal support for ASN.1 DER binary encoding.
 #
@@ -22,13 +22,20 @@
 
 import struct
 
-from Crypto.Util.py3compat import byte_string, b, bchr, bord
+from Crypto.Util.py3compat import byte_string, bchr, bord
 
 from Crypto.Util.number import long_to_bytes, bytes_to_long
 
-__all__ = ['DerObject', 'DerInteger', 'DerOctetString', 'DerNull',
-           'DerSequence', 'DerObjectId', 'DerBitString', 'DerSetOf']
+__all__ = ['DerObject', 'DerInteger', 'DerBoolean', 'DerOctetString',
+           'DerNull', 'DerSequence', 'DerObjectId', 'DerBitString', 'DerSetOf']
 
+# Useful references:
+# - https://luca.ntop.org/Teaching/Appunti/asn1.html
+# - https://letsencrypt.org/docs/a-warm-welcome-to-asn1-and-der/
+# - https://www.zytrax.com/tech/survival/asn1.html
+# - https://www.oss.com/asn1/resources/books-whitepapers-pubs/larmouth-asn1-book.pdf
+# - https://www.itu.int/ITU-T/studygroups/com17/languages/X.690-0207.pdf
+# - https://misc.daniel-marschall.de/asn.1/oid-converter/online.php
 
 def _is_number(x, only_non_negative=False):
     test = 0
@@ -46,7 +53,7 @@ class BytesIO_EOF(object):
     def __init__(self, initial_bytes):
         self._buffer = initial_bytes
         self._index = 0
-        self._bookmark  = None
+        self._bookmark = None
 
     def set_bookmark(self):
         self._bookmark = self._index
@@ -61,7 +68,7 @@ class BytesIO_EOF(object):
     def read(self, length):
         new_index = self._index + length
         if new_index > len(self._buffer):
-            raise ValueError
+            raise ValueError("Not enough data for DER decoding: expected %d bytes and found %d" % (new_index, len(self._buffer)))
 
         result = self._buffer[self._index:new_index]
         self._index = new_index
@@ -77,12 +84,12 @@ class DerObject(object):
         This class should never be directly instantiated.
         """
 
-        def __init__(self, asn1Id=None, payload=b(''), implicit=None,
+        def __init__(self, asn1Id=None, payload=b'', implicit=None,
                      constructed=False, explicit=None):
                 """Initialize the DER object according to a specific ASN.1 type.
 
                 :Parameters:
-                  asn1Id : integer
+                  asn1Id : integer or byte
                     The universal DER tag number for this object
                     (e.g. 0x10 for a SEQUENCE).
                     If None, the tag is not known yet.
@@ -92,16 +99,20 @@ class DerObject(object):
                     the content octets).
                     If not specified, the payload is empty.
 
-                  implicit : integer
-                    The IMPLICIT tag number to use for the encoded object.
+                  implicit : integer or byte
+                    The IMPLICIT tag number (< 0x1F) to use for the encoded object.
                     It overrides the universal tag *asn1Id*.
+                    It cannot be combined with the ``explicit`` parameter.
+                    By default, there is no IMPLICIT tag.
 
                   constructed : bool
                     True when the ASN.1 type is *constructed*.
-                    False when it is *primitive*.
+                    False when it is *primitive* (default).
 
-                  explicit : integer
-                    The EXPLICIT tag number to use for the encoded object.
+                  explicit : integer or byte
+                    The EXPLICIT tag number (< 0x1F) to use for the encoded object.
+                    It cannot be combined with the ``implicit`` parameter.
+                    By default, there is no EXPLICIT tag.
                 """
 
                 if asn1Id is None:
@@ -125,23 +136,26 @@ class DerObject(object):
                 # context-spec |   1      0 (default for IMPLICIT/EXPLICIT)
                 # private      |   1      1
                 #
+
+                constructed_bit = 0x20 if constructed else 0x00
+
                 if None not in (explicit, implicit):
                     raise ValueError("Explicit and implicit tags are"
                                      " mutually exclusive")
 
                 if implicit is not None:
-                    self._tag_octet = 0x80 | 0x20 * constructed | self._convertTag(implicit)
-                    return
-
-                if explicit is not None:
-                    self._tag_octet = 0xA0 | self._convertTag(explicit)
-                    self._inner_tag_octet = 0x20 * constructed | asn1Id
-                    return
-                
-                self._tag_octet = 0x20 * constructed | asn1Id
+                    # IMPLICIT tag overrides asn1Id
+                    self._tag_octet = 0x80 | constructed_bit | self._convertTag(implicit)
+                elif explicit is not None:
+                    # 'constructed bit' is always asserted for an EXPLICIT tag
+                    self._tag_octet = 0x80 | 0x20 | self._convertTag(explicit)
+                    self._inner_tag_octet = constructed_bit | asn1Id
+                else:
+                    # Neither IMPLICIT nor EXPLICIT
+                    self._tag_octet = constructed_bit | asn1Id
 
         def _convertTag(self, tag):
-                """Check if *tag* is a real DER tag.
+                """Check if *tag* is a real DER tag (5 bits).
                 Convert it from a character to number if necessary.
                 """
                 if not _is_number(tag):
@@ -185,14 +199,16 @@ class DerObject(object):
                 """Decode DER length octets from a file."""
 
                 length = s.read_byte()
-                if length <= 127:
-                        return length
-                payloadLength = bytes_to_long(s.read(length & 0x7F))
-                # According to DER (but not BER) the long form is used
-                # only when the length doesn't fit into 7 bits.
-                if payloadLength <= 127:
-                        raise ValueError("Not a DER length tag (but still valid BER).")
-                return payloadLength
+
+                if length > 127:
+                    encoded_length = s.read(length & 0x7F)
+                    if bord(encoded_length[0]) == 0:
+                        raise ValueError("Invalid DER: length has leading zero")
+                    length = bytes_to_long(encoded_length)
+                    if length <= 127:
+                        raise ValueError("Invalid DER: length in long form but smaller than 128")
+
+                return length
 
         def decode(self, der_encoded, strict=False):
                 """Decode a complete DER element, and re-initializes this
@@ -284,7 +300,7 @@ class DerInteger(DerObject):
                     It overrides the universal tag for INTEGER (2).
                 """
 
-                DerObject.__init__(self, 0x02, b(''), implicit,
+                DerObject.__init__(self, 0x02, b'', implicit,
                                    False, explicit)
                 self.value = value  # The integer value
 
@@ -293,7 +309,7 @@ class DerInteger(DerObject):
                 binary string."""
 
                 number = self.value
-                self.payload = b('')
+                self.payload = b''
                 while True:
                     self.payload = bchr(int(number & 255)) + self.payload
                     if 128 <= number <= 255:
@@ -304,7 +320,7 @@ class DerInteger(DerObject):
                 return DerObject.encode(self)
 
         def decode(self, der_encoded, strict=False):
-                """Decode a complete DER INTEGER DER, and re-initializes this
+                """Decode a DER-encoded INTEGER, and re-initializes this
                 object with it.
 
                 Args:
@@ -339,6 +355,89 @@ class DerInteger(DerObject):
                     self.value -= bits
 
 
+class DerBoolean(DerObject):
+    """Class to model a DER-encoded BOOLEAN.
+
+    An example of encoding is::
+
+    >>> from Crypto.Util.asn1 import DerBoolean
+    >>> bool_der = DerBoolean(True)
+    >>> print(bool_der.encode().hex())
+
+    which will show ``0101ff``, the DER encoding of True.
+
+    And for decoding::
+
+    >>> s = bytes.fromhex('0101ff')
+    >>> try:
+    >>>   bool_der = DerBoolean()
+    >>>   bool_der.decode(s)
+    >>>   print(bool_der.value)
+    >>> except ValueError:
+    >>>   print "Not a valid DER BOOLEAN"
+
+    the output will be ``True``.
+
+    :ivar value: The boolean value
+    :vartype value: boolean
+    """
+    def __init__(self, value=False, implicit=None, explicit=None):
+        """Initialize the DER object as a BOOLEAN.
+
+        Args:
+          value (boolean):
+            The value of the boolean. Default is False.
+
+          implicit (integer or byte):
+            The IMPLICIT tag number (< 0x1F) to use for the encoded object.
+            It overrides the universal tag for BOOLEAN (1).
+            It cannot be combined with the ``explicit`` parameter.
+            By default, there is no IMPLICIT tag.
+
+          explicit (integer or byte):
+            The EXPLICIT tag number (< 0x1F) to use for the encoded object.
+            It cannot be combined with the ``implicit`` parameter.
+            By default, there is no EXPLICIT tag.
+        """
+
+        DerObject.__init__(self, 0x01, b'', implicit, False, explicit)
+        self.value = value  # The boolean value
+
+    def encode(self):
+        """Return the DER BOOLEAN, fully encoded as a binary string."""
+
+        self.payload = b'\xFF' if self.value else b'\x00'
+        return DerObject.encode(self)
+
+    def decode(self, der_encoded, strict=False):
+        """Decode a DER-encoded BOOLEAN, and re-initializes this object with it.
+
+        Args:
+            der_encoded (byte string): A DER-encoded BOOLEAN.
+
+        Raises:
+            ValueError: in case of parsing errors.
+        """
+
+        return DerObject.decode(self, der_encoded, strict)
+
+    def _decodeFromStream(self, s, strict):
+        """Decode a DER-encoded BOOLEAN from a file."""
+
+        # Fill up self.payload
+        DerObject._decodeFromStream(self, s, strict)
+
+        if len(self.payload) != 1:
+            raise ValueError("Invalid encoding for DER BOOLEAN: payload is not 1 byte")
+
+        if bord(self.payload[0]) == 0:
+            self.value = False
+        elif bord(self.payload[0]) == 0xFF:
+            self.value = True
+        else:
+            raise ValueError("Invalid payload for DER BOOLEAN")
+
+
 class DerSequence(DerObject):
         """Class to model a DER SEQUENCE.
 
@@ -382,7 +481,7 @@ class DerSequence(DerObject):
 
         """
 
-        def __init__(self, startSeq=None, implicit=None):
+        def __init__(self, startSeq=None, implicit=None, explicit=None):
                 """Initialize the DER object as a SEQUENCE.
 
                 :Parameters:
@@ -390,12 +489,19 @@ class DerSequence(DerObject):
                     A sequence whose element are either integers or
                     other DER objects.
 
-                  implicit : integer
-                    The IMPLICIT tag to use for the encoded object.
+                  implicit : integer or byte
+                    The IMPLICIT tag number (< 0x1F) to use for the encoded object.
                     It overrides the universal tag for SEQUENCE (16).
+                    It cannot be combined with the ``explicit`` parameter.
+                    By default, there is no IMPLICIT tag.
+
+                  explicit : integer or byte
+                    The EXPLICIT tag number (< 0x1F) to use for the encoded object.
+                    It cannot be combined with the ``implicit`` parameter.
+                    By default, there is no EXPLICIT tag.
                 """
 
-                DerObject.__init__(self, 0x10, b(''), implicit, True)
+                DerObject.__init__(self, 0x10, b'', implicit, True, explicit)
                 if startSeq is None:
                     self._seq = []
                 else:
@@ -432,6 +538,10 @@ class DerSequence(DerObject):
                 self._seq.append(item)
                 return self
 
+        def insert(self, index, item):
+                self._seq.insert(index, item)
+                return self
+
         def hasInts(self, only_non_negative=True):
                 """Return the number of items in this sequence that are
                 integers.
@@ -440,9 +550,9 @@ class DerSequence(DerObject):
                   only_non_negative (boolean):
                     If ``True``, negative integers are not counted in.
                 """
-                def _is_number2(x):
-                    return _is_number(x, only_non_negative)
-                return len(list(filter(_is_number2, self._seq)))
+
+                items = [x for x in self._seq if _is_number(x, only_non_negative)]
+                return len(items)
 
         def hasOnlyInts(self, only_non_negative=True):
                 """Return ``True`` if all items in this sequence are integers
@@ -465,7 +575,7 @@ class DerSequence(DerObject):
                   ValueError: if some elements in the sequence are neither integers
                               nor byte strings.
                 """
-                self.payload = b('')
+                self.payload = b''
                 for item in self._seq:
                     if byte_string(item):
                         self.payload += item
@@ -475,7 +585,7 @@ class DerSequence(DerObject):
                         self.payload += item.encode()
                 return DerObject.encode(self)
 
-        def decode(self, der_encoded, nr_elements=None, only_ints_expected=False, strict=False):
+        def decode(self, der_encoded, strict=False, nr_elements=None, only_ints_expected=False):
                 """Decode a complete DER SEQUENCE, and re-initializes this
                 object with it.
 
@@ -486,7 +596,7 @@ class DerSequence(DerObject):
                     The number of members the SEQUENCE can have
                   only_ints_expected (boolean):
                     Whether the SEQUENCE is expected to contain only integers.
-                  struct (boolean):
+                  strict (boolean):
                     Whether decoding must check for strict DER compliancy.
 
                 Raises:
@@ -525,7 +635,6 @@ class DerSequence(DerObject):
                         self._seq.append(p.data_since_bookmark())
                     else:
                         derInt = DerInteger()
-                        #import pdb; pdb.set_trace()
                         data = p.data_since_bookmark()
                         derInt.decode(data, strict=strict)
                         self._seq.append(derInt.value)
@@ -572,7 +681,7 @@ class DerOctetString(DerObject):
     :vartype payload: byte string
     """
 
-    def __init__(self, value=b(''), implicit=None):
+    def __init__(self, value=b'', implicit=None):
         """Initialize the DER object as an OCTET STRING.
 
         :Parameters:
@@ -593,7 +702,7 @@ class DerNull(DerObject):
     def __init__(self):
         """Initialize the DER object as a NULL."""
 
-        DerObject.__init__(self, 0x05, b(''), None, False)
+        DerObject.__init__(self, 0x05, b'', None, False)
 
 
 class DerObjectId(DerObject):
@@ -638,39 +747,50 @@ class DerObjectId(DerObject):
           explicit : integer
             The EXPLICIT tag to use for the encoded object.
         """
-        DerObject.__init__(self, 0x06, b(''), implicit, False, explicit)
+        DerObject.__init__(self, 0x06, b'', implicit, False, explicit)
         self.value = value
 
     def encode(self):
         """Return the DER OBJECT ID, fully encoded as a
         binary string."""
 
-        comps = list(map(int, self.value.split(".")))
+        comps = [int(x) for x in self.value.split(".")]
+
         if len(comps) < 2:
             raise ValueError("Not a valid Object Identifier string")
-        self.payload = bchr(40*comps[0]+comps[1])
-        for v in comps[2:]:
-            enc = []
+        if comps[0] > 2:
+            raise ValueError("First component must be 0, 1 or 2")
+        if comps[0] < 2 and comps[1] > 39:
+            raise ValueError("Second component must be 39 at most")
+
+        subcomps = [40 * comps[0] + comps[1]] + comps[2:]
+
+        encoding = []
+        for v in reversed(subcomps):
+            encoding.append(v & 0x7F)
+            v >>= 7
             while v:
-                enc.insert(0, (v & 0x7F) | 0x80)
+                encoding.append((v & 0x7F) | 0x80)
                 v >>= 7
-            enc[-1] &= 0x7F
-            self.payload += b('').join(map(bchr, enc))
+
+        self.payload = b''.join([bchr(x) for x in reversed(encoding)])
         return DerObject.encode(self)
 
-    def decode(self, der_encoded):
+    def decode(self, der_encoded, strict=False):
         """Decode a complete DER OBJECT ID, and re-initializes this
         object with it.
 
         Args:
             der_encoded (byte string):
                 A complete DER OBJECT ID.
+            strict (boolean):
+                Whether decoding must check for strict DER compliancy.
 
         Raises:
             ValueError: in case of parsing errors.
         """
 
-        return DerObject.decode(self, der_encoded)
+        return DerObject.decode(self, der_encoded, strict)
 
     def _decodeFromStream(self, s, strict):
         """Decode a complete DER OBJECT ID from a file."""
@@ -680,15 +800,27 @@ class DerObjectId(DerObject):
 
         # Derive self.value from self.payload
         p = BytesIO_EOF(self.payload)
-        comps = list(map(str, divmod(p.read_byte(), 40)))
+
+        subcomps = []
         v = 0
         while p.remaining_data():
             c = p.read_byte()
-            v = v*128 + (c & 0x7F)
+            v = (v << 7) + (c & 0x7F)
             if not (c & 0x80):
-                comps.append(str(v))
+                subcomps.append(v)
                 v = 0
-        self.value = '.'.join(comps)
+
+        if len(subcomps) == 0:
+            raise ValueError("Empty payload")
+
+        if subcomps[0] < 40:
+            subcomps[:1] = [0, subcomps[0]]
+        elif subcomps[0] < 80:
+            subcomps[:1] = [1, subcomps[0] - 40]
+        else:
+            subcomps[:1] = [2, subcomps[0] - 80]
+
+        self.value = ".".join([str(x) for x in subcomps])
 
 
 class DerBitString(DerObject):
@@ -697,21 +829,20 @@ class DerBitString(DerObject):
     An example of encoding is:
 
     >>> from Crypto.Util.asn1 import DerBitString
-    >>> from binascii import hexlify, unhexlify
-    >>> bs_der = DerBitString(b'\\xaa')
-    >>> bs_der.value += b'\\xbb'
-    >>> print hexlify(bs_der.encode())
+    >>> bs_der = DerBitString(b'\\xAA')
+    >>> bs_der.value += b'\\xBB'
+    >>> print(bs_der.encode().hex())
 
-    which will show ``040300aabb``, the DER encoding for the bit string
+    which will show ``030300aabb``, the DER encoding for the bit string
     ``b'\\xAA\\xBB'``.
 
     For decoding:
 
-    >>> s = unhexlify(b'040300aabb')
+    >>> s = bytes.fromhex('030300aabb')
     >>> try:
     >>>   bs_der = DerBitString()
     >>>   bs_der.decode(s)
-    >>>   print hexlify(bs_der.value)
+    >>>   print(bs_der.value.hex())
     >>> except ValueError:
     >>>   print "Not a valid DER BIT STRING"
 
@@ -721,7 +852,7 @@ class DerBitString(DerObject):
     :vartype value: byte string
     """
 
-    def __init__(self, value=b(''), implicit=None, explicit=None):
+    def __init__(self, value=b'', implicit=None, explicit=None):
         """Initialize the DER object as a BIT STRING.
 
         :Parameters:
@@ -730,11 +861,11 @@ class DerBitString(DerObject):
             If not specified, the bit string is empty.
           implicit : integer
             The IMPLICIT tag to use for the encoded object.
-            It overrides the universal tag for OCTET STRING (3).
+            It overrides the universal tag for BIT STRING (3).
           explicit : integer
             The EXPLICIT tag to use for the encoded object.
         """
-        DerObject.__init__(self, 0x03, b(''), implicit, False, explicit)
+        DerObject.__init__(self, 0x03, b'', implicit, False, explicit)
 
         # The bitstring value (packed)
         if isinstance(value, DerObject):
@@ -744,24 +875,26 @@ class DerBitString(DerObject):
 
     def encode(self):
         """Return the DER BIT STRING, fully encoded as a
-        binary string."""
+        byte string."""
 
         # Add padding count byte
-        self.payload = b('\x00') + self.value
+        self.payload = b'\x00' + self.value
         return DerObject.encode(self)
 
-    def decode(self, der_encoded):
+    def decode(self, der_encoded, strict=False):
         """Decode a complete DER BIT STRING, and re-initializes this
         object with it.
 
         Args:
             der_encoded (byte string): a complete DER BIT STRING.
+            strict (boolean):
+                Whether decoding must check for strict DER compliancy.
 
         Raises:
             ValueError: in case of parsing errors.
         """
 
-        return DerObject.decode(self, der_encoded)
+        return DerObject.decode(self, der_encoded, strict)
 
     def _decodeFromStream(self, s, strict):
         """Decode a complete DER BIT STRING DER from a file."""
@@ -773,7 +906,7 @@ class DerBitString(DerObject):
             raise ValueError("Not a valid BIT STRING")
 
         # Fill-up self.value
-        self.value = b('')
+        self.value = b''
         # Remove padding count byte
         if self.payload:
             self.value = self.payload[1:]
@@ -816,7 +949,7 @@ class DerSetOf(DerObject):
             The IMPLICIT tag to use for the encoded object.
             It overrides the universal tag for SET OF (17).
         """
-        DerObject.__init__(self, 0x11, b(''), implicit, True)
+        DerObject.__init__(self, 0x11, b'', implicit, True)
         self._seq = []
 
         # All elements must be of the same type (and therefore have the
@@ -860,7 +993,7 @@ class DerSetOf(DerObject):
         if elem not in self._seq:
             self._seq.append(elem)
 
-    def decode(self, der_encoded):
+    def decode(self, der_encoded, strict=False):
         """Decode a complete SET OF DER element, and re-initializes this
         object with it.
 
@@ -869,12 +1002,14 @@ class DerSetOf(DerObject):
 
         Args:
             der_encoded (byte string): a complete DER BIT SET OF.
+            strict (boolean):
+                Whether decoding must check for strict DER compliancy.
 
         Raises:
             ValueError: in case of parsing errors.
         """
 
-        return DerObject.decode(self, der_encoded)
+        return DerObject.decode(self, der_encoded, strict)
 
     def _decodeFromStream(self, s, strict):
         """Decode a complete DER SET OF from a file."""
@@ -925,5 +1060,5 @@ class DerSetOf(DerObject):
                 bys = item
             ordered.append(bys)
         ordered.sort()
-        self.payload = b('').join(ordered)
+        self.payload = b''.join(ordered)
         return DerObject.encode(self)
